1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
160
161
162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
177
178
179
180
181
182
183
184
185
186
187
188
189
190
191
192
193
194
195
196
197
198
199
200
201
202
203
204
205
206
207
208
209
210
211
212
213
214
215
216
217
218
219
220
221
222
223
224
225
226
227
228
229
230
231
232
233
234
235
236
237
238
239
240
241
242
243
244
245
246
247
248
249
250
251
252
253
254
255
256
257
258
259
260
261
262
263
264
265
266
267
268
269
270
271
272
273
274
275
276
277
278
279
280
281
282
283
284
285
286
287
288
289
290
291
292
293
294
295
296
297
298
299
300
301
302
303
304
305
306
307
308
309
310
311
312
313
314
315
316
317
318
319
320
321
322
323
324
325
326
327
328
329
330
331
332
333
334
335
336
337
338
339
340
341
342
343
344
345
346
347
348
349
350
351
352
353
354
355
356
357
358
359
360
361
362
363
364
365
366
367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
394
395
396
397
398
399
400
401
402
403
404
405
406
407
408
409
410
411
412
413
414
415
416
417
418
419
420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
448
449
450
451
452
453
454
455
456
457
458
459
460
461
462
463
464
465
466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
505
506
507
508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
573
574
575
576
577
578
579
580
581
582
583
584
585
586
587
588
589
590
591
592
593
594
595
596
597
598
599
600
601
602
603
604
605
606
607
608
609
610
611
612
613
614
615
616
617
618
619
620
621
622
623
624
625
626
627
628
629
630
631
632
633
634
635
636
637
638
639
640
641
642
643
644
645
646
647
648
649
650
651
652
653
654
655
656
657
658
659
660
661
662
663
664
665
666
667
668
669
670
671
672
673
674
675
676
677
678
679
680
681
682
683
684
685
686
687
688
689
690
691
692
693
694
695
696
697
698
699
700
701
702
703
704
705
706
707
708
709
710
711
712
713
714
715
716
717
718
719
720
721
722
723
724
725
726
727
728
729
730
731
732
733
734
735
736
737
738
739
740
741
742
743
744
745
746
747
748
749
750
751
752
753
754
755
756
757
758
759
760
761
762
763
764
765
766
767
768
769
770
771
772
773
774
775
776
777
778
779
780
781
782
783
784
785
786
787
788
789
790
791
792
793
794
795
796
797
798
799
800
801
802
803
804
805
806
807
808
809
810
811
812
813
814
815
816
817
818
819
820
821
822
823
824
825
826
827
828
829
830
831
832
833
834
835
836
837
838
839
840
841
842
843
844
845
846
847
848
849
850
851
852
853
854
855
856
857
858
859
860
861
862
863
864
865
866
867
868
869
870
871
872
873
874
875
876
877
878
879
880
881
882
883
884
885
886
887
888
889
890
891
892
893
894
895
896
897
898
899
900
901
902
903
904
905
906
907
908
909
910
911
912
913
914
915
916
917
918
919
920
921
922
923
924
925
926
927
928
929
930
931
932
933
934
935
936
937
938
939
940
941
942
943
944
945
946
947
948
949
950
951
952
953
954
955
956
957
958
959
960
961
962
963
964
965
966
967
968
969
970
971
972
973
974
975
976
977
978
979
980
981
982
983
984
985
986
987
988
989
990
991
992
993
994
995
996
997
998
999
1000
1001
1002
1003
1004
1005
1006
1007
1008
1009
1010
1011
1012
1013
1014
1015
1016
1017
1018
1019
1020
1021
1022
1023
1024
1025
1026
1027
1028
1029
1030
1031
1032
1033
1034
1035
1036
1037
1038
1039
1040
1041
1042
1043
1044
1045
1046
1047
1048
1049
1050
1051
1052
1053
1054
1055
1056
1057
1058
1059
1060
1061
1062
1063
1064
1065
1066
1067
1068
1069
1070 | """
Metaprogramming subroutines (using reikna, pyopencl, pycuda)
========================================================
"""
from __future__ import absolute_import # Python2 compatibility
# from . import cSelect
def create_kernel_sets(API):
"""
Create the kernel from the kernel sets.
Note that in some tests (Benoit's and my tests) CUDA shows some degraded accuracy.
This loss of accuracy was due to undefined shared memory behavior, which I don't fully understand.
This has been fixed in 2019.2.0 as the operations are moved to global memory.
"""
kernel_sets = ( cMultiplyScalar() +
cCopy() +
cTensorCopy() +
cTensorMultiply() +
cMultiplyVec() +
cHypot() +
cAddScalar() +
cSelect() +
cAddVec() +
cMultiplyVecInplace() +
cMultiplyConjVecInplace() +
cMultiplyRealInplace() +
cMultiplyConjVec() +
cDiff() +
cSqrt() +
cAnisoShrink() +
cSpmv() +
cSpmvh() +
cHadamard())
# if 'cuda' is API:
# print('Select cuda interface')
# kernel_sets = atomic_add.cuda_add + kernel_sets
# elif 'ocl' is API:
# print("Selecting opencl interface")
# kernel_sets = atomic_add.ocl_add + kernel_sets
kernel_sets = atomic_add(API) + kernel_sets
return kernel_sets
def cMultiplyScalar():
"""
Return the kernel source for cMultiplyScalar.
"""
code_text ="""
KERNEL void cMultiplyScalar(
const float2 CA,
GLOBAL_MEM float2 *CX)
{
// Scale CX by CA: CX=CA*CX
// CA: scaling factor(float2)
//*CX: input, output array(float2)
unsigned long gid = get_global_id(0);
CX[gid].x=CA.x*CX[gid].x-CA.y*CX[gid].y;
CX[gid].y=CA.x*CX[gid].y+CA.y*CX[gid].x;
};
"""
return code_text
def cCopy():
"""
Return the kernel source for cCopy
"""
code_text = """
KERNEL void cCopy(
GLOBAL_MEM const float2 *CX,
GLOBAL_MEM float2 *CY)
{
// Copy x to y: y = x;
//CX: input array (float2)
// CY output array (float2)
unsigned long gid=get_global_id(0);
CY[gid]=CX[gid];
};
"""
return code_text
def cTensorCopy():
"""
Return the kernel source for cTensorCopy.
"""
code_text = """
KERNEL void cTensorCopy(
const unsigned int batch,
const unsigned int dim,
GLOBAL_MEM const unsigned int *Nd_elements,
GLOBAL_MEM const unsigned int *Kd_elements,
GLOBAL_MEM const float *invNd,
GLOBAL_MEM const float2 *indata,
GLOBAL_MEM float2 *outdata,
const int direction)
{
const unsigned int gid=get_global_id(0);
unsigned int curr_res = gid;
unsigned int new_idx = 0;
unsigned int group;
for (unsigned int dimid =0; dimid < dim; dimid ++){
group = (float)curr_res*invNd[dimid];
new_idx += group * Kd_elements[dimid];
curr_res = curr_res - group * Nd_elements[dimid];
};
if (direction == 1) {
for (unsigned int bat=0; bat < batch; bat ++ )
{
outdata[new_idx*batch+bat]= indata[gid*batch+bat];
};
};
if (direction == -1) {
for (unsigned int bat=0; bat < batch; bat ++ )
{
outdata[gid*batch+bat]= indata[new_idx*batch+bat];
};
};
};
"""
return code_text
def cTensorMultiply():
"""
Return the kernel source for cTensorMultiply
"""
code_text = """
KERNEL void cTensorMultiply(
const unsigned int batch, // batch
const unsigned int dim, // dimensions
GLOBAL_MEM const unsigned int *Nd, // In batch mode, Nd*batch but is calculated outside the kernel
GLOBAL_MEM const unsigned int *Nd_elements, // Number of elements to move along the dimension = strides / itemsize
GLOBAL_MEM const float *invNd_elements, // float: inverse of the Nd_elements, which aims for fast division // batch mode: Nd_elements / batch
GLOBAL_MEM const float *vec, // Real, vector, length sum Nd[dimid]
GLOBAL_MEM float2 *outdata,
const unsigned int div)
{
const unsigned int gid=get_global_id(0);
const unsigned int pid = (float)gid / (float)batch;
// const unsigned int bat = gid - pid * batch;
unsigned int group;
unsigned int Nd_indx_shift = 0;
float mul = 1.0;
unsigned int res = pid;
for (unsigned int dimid = 0; dimid < dim; dimid ++){
group = (float)res * invNd_elements[dimid]; // The index along the axis
res = res - group * Nd_elements[dimid];
const unsigned int N = Nd[dimid];
mul = mul * vec[group + Nd_indx_shift];
Nd_indx_shift = Nd_indx_shift + N;
}
if (div == 1){
// for (unsigned int bat = 0; bat < batch; bat ++ )
// {
float2 tmp = outdata[gid];
tmp.x = tmp.x / mul;
tmp.y = tmp.y / mul;
outdata[gid] = tmp;
// };
};
if (div == 0){
// for (unsigned int bat = 0; bat < batch; bat ++ )
// {
float2 tmp = outdata[gid];
tmp.x = tmp.x * mul;
tmp.y = tmp.y * mul;
outdata[gid] = tmp;
// };
};
};
"""
return code_text
def atomic_add(API):
"""
Return the atomic_add for the given API.
Overcome the missing atomic_add_float for OpenCL-1.2.
Note: will be checked if OpenCL 2.0 provided by all GPU vendors.
"""
ocl_add = """
// #pragma OPENCL EXTENSION cl_khr_int64_base_atomics : enable
KERNEL void atomic_add_float(
GLOBAL_MEM float *ptr,
const float temp)
{
// The work-around of AtomicAdd for float
// lockless add *source += operand
// Caution!!!!!!! Use with care! You have been warned!
// http://simpleopencl.blogspot.com/2013/05/atomic-operations-and-floats-in-opencl.html
// Source: https://github.com/clMathLibraries/clSPARSE/blob/master/src/library/kernels/csrmv_adaptive.cl
union {
unsigned int intVal;
float floatVal;
} newVal;
union {
unsigned int intVal;
float floatVal;
} prevVal;
do {
prevVal.floatVal = *ptr;
newVal.floatVal = prevVal.floatVal + temp;
} while (atomic_cmpxchg((__global volatile unsigned int *)ptr, prevVal.intVal, newVal.intVal) != prevVal.intVal);
};
KERNEL void atomic_add_float2_old(
GLOBAL_MEM float2 *ptr,
const float2 temp)
{
atomic_add_float(
(GLOBAL_MEM float*)ptr,
temp.x);
atomic_add_float(
(GLOBAL_MEM float*)ptr+1,
temp.y);
};
KERNEL void atomic_add_float2(
GLOBAL_MEM float2 *ptr,
const float2 temp,
GLOBAL_MEM float2 *res)
{ // Add atomic two sum for real/imaginary parts
// *res array stores the error
union {
unsigned int intVal;
float floatVal;
} newVal;
union {
unsigned int intVal;
float floatVal;
} prevVal;
float x;
float val;
float newacc;
float r;
__global const float * realptr = ( __global float *)ptr;
__global const float * imagptr = ( __global float *)ptr + 1;
__global const float * realres = ( __global float *)res;
__global const float * imagres = ( __global float *)res + 1;
do { // atomic add of the value
prevVal.floatVal = *realptr ;
newVal.floatVal = prevVal.floatVal + temp.x;
} while (atomic_cmpxchg((__global volatile unsigned int *)realptr, prevVal.intVal, newVal.intVal) != prevVal.intVal);
x = prevVal.floatVal;
val = temp.x;
if (fabs(x) < fabs(val))
{
const float swap = x;
x = val;
val = swap;
}
newacc = x + val;
r = val - (newacc - x); // Recover rounding error using Fast2Sum, assumes |oldacc|>=|val|
do { // atomic_add(realres, r)
prevVal.floatVal = *realres ;
newVal.floatVal = prevVal.floatVal + r;
} while (atomic_cmpxchg((__global volatile unsigned int *)realres, prevVal.intVal, newVal.intVal) != prevVal.intVal);
// End of real part
// Now do the imaginary part
do {
prevVal.floatVal = *imagptr;
newVal.floatVal = prevVal.floatVal + temp.y;
} while (atomic_cmpxchg((__global volatile unsigned int *)imagptr, prevVal.intVal, newVal.intVal) != prevVal.intVal);
x = prevVal.floatVal; // this copy the old value to x
val = temp.y; // this copy the real part to val
if (fabs(x) < fabs(val))
{
const float swap = x;
x = val;
val = swap;
}
newacc = x + val;
r = val - (newacc - x); // Recover rounding error using Fast2Sum, assumes |oldacc|>=|val|
do {
prevVal.floatVal = *imagres ;
newVal.floatVal = prevVal.floatVal + r;
} while (atomic_cmpxchg((__global volatile unsigned int *)imagres, prevVal.intVal, newVal.intVal) != prevVal.intVal);
};
"""
cuda_add = """
__device__ void atomic_add_float_twosum(
GLOBAL_MEM float * address,
float val,
GLOBAL_MEM float * residue)
{
// https://devtalk.nvidia.com/default/topic/817899/atomicadd-kahan-summation/
// By Sylvain Collange
// Also add the branch clsparse
float x = atomicAdd(address, val);
if (fabs(x) < fabs(val))
{
const float swap = x;
x = val;
val = swap;
}
float newacc = x + val;
float r = val - (newacc - x); // Recover rounding error using Fast2Sum, assumes |oldacc|>=|val|
atomicAdd(residue, r); // Accumulate error in residue
}
__device__ void atomic_add_float(
GLOBAL_MEM float *ptr,
const float temp)
{ // Wrapper around CUDA atomicAdd();
atomicAdd(ptr, temp);
};
__device__ void atomic_add_float2(
GLOBAL_MEM float2 * ptr,
const float2 temp,
GLOBAL_MEM float2 *res)
{ // Wrapper around CUDA atomicAdd();
// atomicAdd((float*)ptr, temp.x);
// atomicAdd((float*)ptr+1, temp.y);
atomic_add_float_twosum((float*)ptr, temp.x, (float*)res);
atomic_add_float_twosum((float*)ptr+1, temp.y, (float*)res+1);
};
"""
if API == 'cuda':
code_text = cuda_add
elif API == 'ocl':
code_text = ocl_add
return code_text
def cMultiplyVec():
"""
Return the kernel source of cMultiplyVec
"""
code_text="""
KERNEL void cMultiplyVec( GLOBAL_MEM float2 *a,
GLOBAL_MEM float2 *b,
GLOBAL_MEM float2 *dest)
{ const unsigned int i = get_global_id(0);
dest[i].x = a[i].x*b[i].x-a[i].y*b[i].y;
dest[i].y = a[i].x*b[i].y+a[i].y*b[i].x;
//barrier(CLK_GLOBAL_MEM_FENCE);
};
"""
return code_text
def cAddScalar():
"""
Return the kernel source for cAddScalar.
"""
code_text ="""
KERNEL void cAddScalar(const float2 CA,
GLOBAL_MEM float2 *CX)
{
// (single complex) scale x by a: x = x + ca;
// CA: add factor
// CX: input and output array (float2)
int gid = get_global_id(0);
CX[gid].x += CA.x;
CX[gid].y += CA.y;
};
"""
return code_text
def cHadamard():
"""
Return the Hadamard operations related kernel sources.
"""
R="""
// Batched copying indata[order1] to outdata[order2]
// Superceded by cTensorCopy()
// However left here as a general function.
KERNEL void cSelect2(
const unsigned int Reps,
GLOBAL_MEM const unsigned int *order1,
GLOBAL_MEM const unsigned int *order2,
GLOBAL_MEM const float2 *indata,
GLOBAL_MEM float2 *outdata)
{
const unsigned int gid=get_global_id(0);
const unsigned int t = ((float)gid / (float)Reps); // indptr
const unsigned int r = gid - t*Reps; // residue
const unsigned int index2 = order2[t]*Reps + r;
const unsigned int index1 = order1[t]*Reps + r;
outdata[index2]=indata[index1];
};
KERNEL void cDistribute(
const unsigned int Reps,
const unsigned int prodNd,
GLOBAL_MEM const float2 *arr_large, // sensitivity and scaling array, prodNd*Reps
GLOBAL_MEM const float2 *arr_image, // image array, prodNd
GLOBAL_MEM float2 *arr_out // output array, prodNd * Reps
)
{
const unsigned int t = get_global_id(0);
//const unsigned int nd = t/Reps;
const unsigned int nd = ((float)t / (float)Reps);
if (nd < prodNd){
const float2 u = arr_large[t];
const float2 v = arr_image[nd];
float2 w;
w.x = u.x * v.x - u.y * v.y;
w.y = u.x * v.y + u.y * v.x;
arr_out[t] = w;
}
}; // End of cDistribute
KERNEL void cMerge(
const unsigned int Reps,
const unsigned int prodNd,
GLOBAL_MEM const float2 *arr_large, // sensitivity and scaling array, prodNd*Reps
GLOBAL_MEM const float2 *arr_image, // image array, prodNd*Reps
GLOBAL_MEM float2 *arr_out // reduced output array, prodNd
)
{
const unsigned int t = get_local_id(0);
//const float float_reps = (float)Reps;
const unsigned int vecWidth=${LL};
// Thread ID within wavefront
const unsigned int id = t & (vecWidth-1);
// One row per wavefront
unsigned int vecsPerBlock=get_local_size(0)/vecWidth;
unsigned int myRow=(get_group_id(0)*vecsPerBlock) + (t/ vecWidth);
LOCAL_MEM float2 partialSums[${LL}];
float2 zero;
zero.x = 0.0;
zero.y = 0.0;
partialSums[t] = zero;
// float2 y= zero;
if (myRow < prodNd)
{
const unsigned int vecStart = myRow * Reps;
const unsigned int vecEnd = vecStart + Reps;
for (unsigned int j = vecStart+id; j<vecEnd; j += vecWidth)
{
const float2 u = arr_large[j]; // sensitivities and scaling, complex
const float2 v = arr_image[j];
float2 w;
w.x = u.x * v.x + u.y * v.y; // conjugate of u
w.y = u.x * v.y - u.y * v.x;
w.x = w.x/(float)Reps;
w.y = w.y/(float)Reps;
partialSums[t] = w; //partialSums[t] + y;
}
LOCAL_BARRIER;
//__syncthreads();
//barrier(CLK_LOCAL_MEM_FENCE);
// Reduce partial sums
unsigned int bar = vecWidth / 2;
while(bar > 0)
{
if (id < bar)
partialSums[t] = partialSums[t] + partialSums[t+bar];
//barrier(CLK_LOCAL_MEM_FENCE);
//__syncthreads();
LOCAL_BARRIER;
bar = bar / 2;
}
// Write result
if (id == 0)
{
arr_out[myRow]=partialSums[t];
}
}
}; // End of cMerge
KERNEL void cAggregate(
const unsigned int Reps,
const unsigned int prodNd,
GLOBAL_MEM const float2 *arr_image, // image array, prodNd*Reps
GLOBAL_MEM float2 *arr_out // reduced output array, prodNd
)
{
const unsigned int t = get_local_id(0);
//const float float_reps = (float)Reps;
const unsigned int vecWidth=${LL};
// Thread ID within wavefront
const unsigned int id = t & (vecWidth-1);
// One row per wavefront
unsigned int vecsPerBlock=get_local_size(0)/vecWidth;
unsigned int myRow=(get_group_id(0)*vecsPerBlock) + (t/ vecWidth);
LOCAL_MEM float2 partialSums[${LL}];
float2 zero;
zero.x = 0.0;
zero.y = 0.0;
partialSums[t] = zero;
// float2 y= zero;
if (myRow < prodNd)
{
const unsigned int vecStart = myRow * Reps;
const unsigned int vecEnd = vecStart + Reps;
for (unsigned int j = vecStart+id; j<vecEnd; j += vecWidth)
{
float2 v = arr_image[j];
v.x = v.x/(float)Reps;
v.y = v.y/(float)Reps;
partialSums[t] =partialSums[t] + v; //partialSums[t] + y;
}
LOCAL_BARRIER;
//__syncthreads();
//barrier(CLK_LOCAL_MEM_FENCE);
// Reduce partial sums
unsigned int bar = vecWidth / 2;
while(bar > 0)
{
if (id < bar)
partialSums[t] = partialSums[t] + partialSums[t+bar];
//barrier(CLK_LOCAL_MEM_FENCE);
//__syncthreads();
LOCAL_BARRIER;
bar = bar / 2;
}
// Write result
if (id == 0)
{
arr_out[myRow]=partialSums[t];
}
}
LOCAL_BARRIER;
}; // End of cAggregate
KERNEL void cPopulate(
const unsigned int Reps,
const unsigned int prodNd,
GLOBAL_MEM const float2 *arr_image, // image array, prodNd
GLOBAL_MEM float2 *arr_out // output array, prodNd * Reps
)
{
const unsigned int t = get_global_id(0);
//const unsigned int nd = t/Reps;
const unsigned int nd = ((float)t / (float)Reps);
if (nd < prodNd){
const float2 v = arr_image[nd];
arr_out[t] = v;
}
LOCAL_BARRIER;
}; // End of cPopulate
"""
return R
def cSpmvh():
"""
Return the cSpmvh related kernel source.
Only pELL_spmvh_mCoil is provided for Spmvh.
NUFFT_hsa_legacy reuses the cCSR_spmv() function, which doubles the storage.
"""
R="""
KERNEL void pELL_spmvh_mCoil_old(
const unsigned int Reps, // number of coils
const unsigned int nRow, // number of rows
const unsigned int prodJd, // product of Jd
const unsigned int sumJd, // sum of Jd
const unsigned int dim, // dimensionality
GLOBAL_MEM const unsigned int *Jd, // Jd
// GLOBAL_MEM const unsigned int *curr_sumJd, //
GLOBAL_MEM const unsigned int *meshindex, // meshindex, prodJd * dim
GLOBAL_MEM const unsigned int *kindx, // unmixed column indexes of all dimensions
GLOBAL_MEM const float2 *udata, // interpolation data before Kronecker product
GLOBAL_MEM float *kx,
GLOBAL_MEM float *ky,
GLOBAL_MEM const float2 *input) // y
{
const unsigned int t = get_local_id(0);
const unsigned int d = get_local_size(0);
const unsigned int g = get_group_id(0);
const unsigned int a = g/prodJd;
const unsigned int b = g - (prodJd*a);
const unsigned int myRow = a*d + t;
const unsigned int j = b;
unsigned int m = myRow / Reps;
unsigned int nc = myRow - m * Reps;
// unsigned int myRow= get_global_id(0);
float2 zero;
zero.x = 0.0;
zero.y = 0.0;
if (m< nRow)
{if (nc < Reps)
{
float2 u = zero;
// for (unsigned int j = 0; j < prodJd; j ++)
if (j<prodJd)
{
// now doing the first dimension
unsigned int index_shift = m * sumJd;
// unsigned int tmp_sumJd = 0;
unsigned int J = Jd[0];
unsigned int index = index_shift + meshindex[dim*j + 0];
unsigned int col = kindx[index] ;
float2 spdata = udata[index];
index_shift += J;
for (unsigned int dimid = 1; dimid < dim; dimid ++ )
{
J = Jd[dimid];
index = index_shift + meshindex[dim*j + dimid]; // the index of the partial ELL arrays *kindx and *udata
col += kindx[index];// + 1 ; // the column index of the current j
float tmp_x = spdata.x;
float2 tmp_udata = udata[index];
spdata.x = tmp_x * tmp_udata.x - spdata.y * tmp_udata.y; // the spdata of the current j
spdata.y = tmp_x * tmp_udata.y + spdata.y * tmp_udata.x;
index_shift += J;
}; // Iterate over dimensions 1 -> Nd - 1
float2 ydata=input[myRow]; // kout[col];
u.x = spdata.x*ydata.x + spdata.y*ydata.y;
u.y = - spdata.y*ydata.x + spdata.x*ydata.y;
atomic_add_float(kx + col*Reps + nc, u.x);
atomic_add_float(ky + col*Reps + nc, u.y);
}; // Iterate for (unsigned int j = 0; j < prodJd; j ++)
}; // if (nc < Reps)
}; // if (m < nRow)
}; // End of pELL_spmvh_mCoil_old
KERNEL void pELL_spmvh_mCoil(
const unsigned int Reps, // number of coils
const unsigned int nRow, // number of rows
const unsigned int prodJd, // product of Jd
const unsigned int sumJd, // sum of Jd
const unsigned int dim, // dimensionality
GLOBAL_MEM const unsigned int *Jd, // Jd
// GLOBAL_MEM const unsigned int *curr_sumJd, //
GLOBAL_MEM const unsigned int *meshindex, // meshindex, prodJd * dim
GLOBAL_MEM const unsigned int *kindx, // unmixed column indexes of all dimensions
GLOBAL_MEM const float2 *udata, // interpolation data before Kronecker product
GLOBAL_MEM float2 *k,
GLOBAL_MEM float2 *res,
GLOBAL_MEM const float2 *input) // y
{
unsigned int myRow0= get_global_id(0);
unsigned int myRow= myRow0/(float)Reps;
unsigned int nc = myRow0 - myRow*Reps;
float2 zero;
zero.x = 0.0;
zero.y = 0.0;
if (myRow < nRow){
for (unsigned int j = 0; j < prodJd; j ++){
float2 u = zero;
// now doing the first dimension
unsigned int index_shift = myRow * sumJd;
// unsigned int tmp_sumJd = 0;
unsigned int J = Jd[0];
unsigned int index = index_shift + meshindex[dim*j + 0];
unsigned int col = kindx[index] ;
float2 spdata = udata[index];
index_shift += J;
for (unsigned int dimid = 1; dimid < dim; dimid ++ ){
J = Jd[dimid];
index = index_shift + meshindex[dim*j + dimid]; // the index of the partial ELL arrays *kindx and *udata
col += kindx[index];// + 1 ; // the column index of the current j
float tmp_x = spdata.x;
float2 tmp_udata = udata[index];
spdata.x = tmp_x * tmp_udata.x - spdata.y * tmp_udata.y; // the spdata of the current j
spdata.y = tmp_x * tmp_udata.y + spdata.y * tmp_udata.x;
index_shift += J;
}; // Iterate over dimensions 1 -> Nd - 1
float2 ydata=input[myRow*Reps + nc]; // kout[col];
u.x = spdata.x*ydata.x + spdata.y*ydata.y;
u.y = - spdata.y*ydata.x + spdata.x*ydata.y;
atomic_add_float2(k + col*Reps + nc, u, res + col*Reps + nc);
}; // Iterate for (unsigned int j = 0; j < prodJd; j ++)
}; // if (m < nRow)
}; // End of pELL_spmvh_mCoil
"""
return R
def cHypot():
"""
Return the kernel code for hypot, which computes the sqrt(x*x + y*y) without intermediate overflow.
"""
R="""
KERNEL void cHypot(GLOBAL_MEM float2 *x,
GLOBAL_MEM const float2 *y)
{
const unsigned int gid = get_global_id(0);
float2 tmp_x;
float2 tmp_y;
tmp_x = x[gid];
tmp_y = y[gid];
tmp_x.x = hypot( tmp_x.x, tmp_x.y); // sqrt( tmp_x.x*tmp_x.x + tmp_x.y*tmp_x.y);
tmp_y.x = hypot( tmp_y.x, tmp_y.y); // sqrt( tmp_y.x*tmp_y.x + tmp_y.y*tmp_y.y);
x[gid].x = hypot(tmp_x.x, tmp_y.x);
x[gid].y = 0.0;
};
"""
return R
def cSpmv():
"""
Return the kernel sources for cSpmv related operations,
providing cCSR_spmv_vector and cpELL_spmv_mCoil.
"""
R = """
KERNEL void cCSR_spmv_vector(
const unsigned int Reps, // Number of coils
const unsigned int numRow,
GLOBAL_MEM const unsigned int *rowDelimiters,
GLOBAL_MEM const unsigned int *cols,
GLOBAL_MEM const float2 *val,
GLOBAL_MEM const float2 *vec,
GLOBAL_MEM float2 *out)
{
const unsigned int t = get_local_id(0);
const unsigned int vecWidth=${LL};
// Thread ID within wavefront
const unsigned int id = t & (vecWidth-1);
// One row per wavefront
unsigned int vecsPerBlock=get_local_size(0)/vecWidth;
unsigned int myRow=(get_group_id(0)*vecsPerBlock) + (t/ vecWidth);
unsigned int m = myRow / Reps;
unsigned int nc = myRow - m * Reps;
LOCAL_MEM float2 partialSums[${LL}];
float2 zero;
zero.x = 0.0;
zero.y = 0.0;
partialSums[t] = zero;
float2 y= zero;
if (myRow < numRow * Reps)
{
const unsigned int vecStart = rowDelimiters[m];
const unsigned int vecEnd = rowDelimiters[m+1];
for (unsigned int j = vecStart+id; j<vecEnd; j += vecWidth)
{
const unsigned int col = cols[j];
const float2 spdata=val[j];
const float2 vecdata=vec[col * Reps + nc];
y.x=spdata.x*vecdata.x - spdata.y*vecdata.y;
y.y=spdata.y*vecdata.x + spdata.x*vecdata.y;
partialSums[t] = partialSums[t] + y;
}
LOCAL_BARRIER;
//__syncthreads();
//barrier(CLK_LOCAL_MEM_FENCE);
// Reduce partial sums
unsigned int bar = vecWidth / 2;
while(bar > 0)
{
if (id < bar)
partialSums[t] = partialSums[t] + partialSums[t+bar];
//barrier(CLK_LOCAL_MEM_FENCE);
//__syncthreads();
LOCAL_BARRIER;
bar = bar / 2;
}
// Write result
if (id == 0)
{
out[myRow]=partialSums[t];
}
}
}; // End of cCSR_spmv_vector
KERNEL void pELL_spmv_mCoil(
const unsigned int Reps, // Number of coils
const unsigned int nRow, // number of rows
const unsigned int prodJd, // product of Jd
const unsigned int sumJd, // sum of Jd
const unsigned int dim, // dimensionality
GLOBAL_MEM const unsigned int *Jd, // Jd, length = dim
//GLOBAL_MEM const unsigned int *curr_sumJd, // summation of Jd[0:dimid]
GLOBAL_MEM const unsigned int *meshindex, // meshindex, prodJd * dim
GLOBAL_MEM const unsigned int *kindx, // unmixed column indexes of all dimensions
GLOBAL_MEM const float2 *udata,// interpolation data before Kronecker product
GLOBAL_MEM const float2 *vec, // multi-channel kspace data, prodKd * Reps
GLOBAL_MEM float2 *out) // multi-channel output, nRow * Reps
{
const unsigned int t = get_local_id(0);
const unsigned int vecWidth=${LL};
// Thread ID within wavefront
const unsigned int id = t & (vecWidth-1);
// One row per wavefront
unsigned int vecsPerBlock=get_local_size(0)/vecWidth;
unsigned int myRow=(get_group_id(0)*vecsPerBlock) + (t/ vecWidth); // the myRow-th non-Cartesian sample
unsigned int m = myRow / Reps;
unsigned int nc = myRow - m * Reps;
LOCAL_MEM float2 partialSums[${LL}];
float2 zero;
zero.x = 0.0;
zero.y = 0.0;
partialSums[t] = zero;
if (myRow < nRow * Reps)
{
const unsigned int vecStart = 0;
const unsigned int vecEnd =prodJd;
float2 y;//=zero;
for (unsigned int j = vecStart+id; j<vecEnd; j += vecWidth)
{ // now doing the first dimension
unsigned int J = Jd[0];
unsigned int index_shift = m * sumJd ;
unsigned int index = index_shift + meshindex[dim*j + 0];
unsigned int col = kindx[index] ;
float2 spdata = udata[index];
index_shift += J;
for (unsigned int dimid = 1; dimid < dim; dimid ++ )
{
unsigned int J = Jd[dimid];
unsigned int index = index_shift + meshindex[dim*j + dimid]; // the index of the partial ELL arrays *kindx and *udata
col += kindx[index] ;//+ 1; // the column index of the current j
float tmp_x= spdata.x;
float2 tmp_udata = udata[index];
spdata.x = spdata.x * tmp_udata.x - spdata.y * tmp_udata.y; // the spdata of the current j
spdata.y = tmp_x * tmp_udata.y + spdata.y * tmp_udata.x;
index_shift += J;
}
float2 vecdata=vec[col * Reps + nc];
y.x = spdata.x*vecdata.x - spdata.y*vecdata.y;
y.y = spdata.y*vecdata.x + spdata.x*vecdata.y;
partialSums[t] = y + partialSums[t];
}
LOCAL_BARRIER;
// Reduce partial sums
unsigned int bar = vecWidth / 2;
while(bar > 0)
{
if (id < bar)
partialSums[t]= partialSums[t]+partialSums[t+bar];
//barrier(CLK_LOCAL_MEM_FENCE);
//__syncthreads();
LOCAL_BARRIER;
bar = bar / 2;
}
// Write result
if (id == 0)
{
out[myRow]=partialSums[t];
}
}
}; // End of pELL_spmv_mCoil
"""
return R
def cMultiplyConjVecInplace():
"""
Return the kernel source of cMultiplyConjVecInplace
"""
R="""
KERNEL void cMultiplyConjVecInplace(
const unsigned int batch,
GLOBAL_MEM float2 *a,
GLOBAL_MEM float2 *outb)
{
const unsigned int gid = get_global_id(0);
const unsigned int voxel_id = (float)gid / (float)batch;
float2 mul = a[voxel_id]; // taking the conjugate
float2 orig = outb[gid];
float2 tmp;
tmp.x = orig.x * mul.x + orig.y * mul.y;
tmp.y = - orig.x * mul.y + orig.y * mul.x;
outb[gid] = tmp;
};
"""
return R
def cMultiplyConjVec():
"""
Return the kernel source of cMultiplyConjVec.
"""
R="""
KERNEL void cMultiplyConjVec(
GLOBAL_MEM float2 *a,
GLOBAL_MEM float2 *b,
GLOBAL_MEM float2 *dest)
{// dest[i]=conj(a[i]) * b[i]
const unsigned int i=get_global_id(0);
dest[i].x=a[i].x*b[i].x+a[i].y*b[i].y;
dest[i].y=a[i].x*b[i].y-a[i].y*b[i].x;
};
"""
return R
def cMultiplyRealInplace():
"""
Return the kernel source of cMultiplyRealInplace.
"""
R="""
KERNEL void cMultiplyRealInplace(
const unsigned int batch,
GLOBAL_MEM const float *a,
GLOBAL_MEM float2 *outb)
{
const unsigned int gid = get_global_id(0);
// const unsigned int voxel_id = gid / batch;
const unsigned int voxel_id = (float)gid / (float)batch;
float mul = a[voxel_id];
float2 orig = outb[gid];
orig.x=orig.x*mul;
orig.y=orig.y*mul;
outb[gid]=orig;
};
"""
return R
def cMultiplyVecInplace():
"""
Return the kernel source of cMultiplyVecInplace.
"""
R="""
KERNEL void cMultiplyVecInplace(
const unsigned int batch,
GLOBAL_MEM const float2 *a,
GLOBAL_MEM float2 *outb)
{
const unsigned int gid = get_global_id(0);
// const unsigned int voxel_id = gid / batch;
const unsigned int voxel_id = (float)gid / (float)batch;
float2 mul = a[voxel_id];
float2 orig = outb[gid];
float2 tmp;
tmp.x=orig.x*mul.x-orig.y*mul.y;
tmp.y=orig.x*mul.y+orig.y*mul.x;
outb[gid]=tmp;
};
"""
return R
def cAddVec():
"""
Return the kernel source for cAddVec.
"""
R="""
KERNEL void cAddVec(
GLOBAL_MEM float2 *a,
GLOBAL_MEM float2 *b,
GLOBAL_MEM float2 *dest)
{const int i = get_global_id(0);
dest[i]= a[i]+b[i];
};"""
return R
def cDiff():
"""
Return the kernel source of cDiff.
"""
R="""
KERNEL void cDiff( GLOBAL_MEM const int *order2,
GLOBAL_MEM const float2 *indata,
GLOBAL_MEM float2 *outdata)
{
const unsigned int gid = get_global_id(0);
const unsigned int ind = order2[gid];
outdata[gid]=indata[ind]- indata[gid];
};
"""
return R
def cAnisoShrink():
"""
Return the kernel source of cAnisoShrink
"""
R="""
KERNEL void cAnisoShrink(const float2 threshold,
GLOBAL_MEM const float2 *indata,
GLOBAL_MEM float2 *outdata)
{
const unsigned int gid = get_global_id(0);
float2 tmp; // temporay register
tmp = indata[gid];
//float zero = 0.0;
//tmp.x=sign(tmp.x)*max(fabs(tmp.x)-threshold.x, zero);
//tmp.y=sign(tmp.y)*max(fabs(tmp.y)-threshold.y, zero);
tmp.x = (tmp.x > threshold.x)*(tmp.x - threshold.x) ;//+ (tmp.x < - threshold.x)*(tmp.x + threshold.x);
tmp.y = (tmp.y > threshold.x)*(tmp.y - threshold.x) ;//+ (tmp.y < - threshold.x)*(tmp.y + threshold.x);
outdata[gid]=tmp;
};
"""
return R
def cSqrt():
"""
Return the kernel source of cSqrt.
"""
R="""
KERNEL void cSqrt(
GLOBAL_MEM float2 *CX)
{
// Copy x to y: y = x;
//CX: input output array (float2)
int gid=get_global_id(0);
CX[gid].x=sqrt(CX[gid].x);
};
"""
return R
def cSelect():
"""
Return the kernel source of cSelect.
"""
R="""
KERNEL void cSelect(
GLOBAL_MEM const int *order1,
GLOBAL_MEM const int *order2,
GLOBAL_MEM const float2 *indata,
GLOBAL_MEM float2 *outdata)
{
const unsigned int gid=get_global_id(0);
outdata[order2[gid]]=
indata[order1[gid]];
};
"""
return R
|